#include <lib.h>
using uid_t = u32;
using gid_t = u32;
using off_t = s64;
using ino_t = u64;
using dev_t = u64;

struct timespec 
{
    s64 seconds;
    s64 nanoseconds;
};

struct Stat
{
    dev_t dev;
    ino_t ino;
    u64 nlink;

    u32 mode;
    uid_t uid;
    gid_t gid;
    u32 _0;
    dev_t rdev;
    off_t size;
    s64 blksize;
    s64 blocks;

    timespec atim;
    timespec mtim;
    timespec ctim;
    s64 _1[3];
};

extern "C" s32 fstat(s32, Stat*);
extern "C" s32 fork();
extern "C" s32 dup2(s32, s32);
extern "C" s32 execve(const char* path_name, char* const argv[], char* const envp[]);
extern "C" s32 waitpid(s32 pid, int* wstatus, int options);

u64 os_file_size(s32 fd)
{
    Stat stat;
    auto result = fstat(fd, &stat);
    assert(result == 0);
    return (u64)stat.size;
}

enum class ExecuteStandardStreamPolicy : u8
{
    inherit,
    pipe,
    ignore,
};

global_variable constexpr u64 standard_stream_count = 2;

struct ExecuteOptions
{
    ExecuteStandardStreamPolicy policies[standard_stream_count]; // INDICES: stdout = 0, stderr = 1
    s32 null_file_descriptor = -1;
};

enum class TerminationKind : u8
{
    unknown,
    exit,
    signal,
    stop,
};

struct Execution
{
    String stdout;
    String stderr;
    TerminationKind termination_kind;
    u32 termination_code;
};

Execution os_execute(Arena* arena, Slice<char* const> arguments, Slice<char* const> environment, ExecuteOptions options)
{
    unused(arena);
    assert(arguments.pointer[arguments.length] == 0);
    assert(environment.pointer[arguments.length] == 0);

    Execution execution = {};

    s32 null_file_descriptor = -1;
    if (options.null_file_descriptor >= 0)
    {
        null_file_descriptor = options.null_file_descriptor;
    }
    else if (options.policies[0] == ExecuteStandardStreamPolicy::ignore || options.policies[1] == ExecuteStandardStreamPolicy::ignore)
    {
        trap();
    }

    int pipes[standard_stream_count][2];

    for (int i = 0; i < 2; i += 1)
    {
        if (options.policies[i] == ExecuteStandardStreamPolicy::pipe)
        {
            trap();
        }
    }

    auto pid = fork();

    switch (pid)
    {
        case -1:
            {
                trap();
            } break;
        case 0: // Child process
            {
                for (u64 i = 0; i < standard_stream_count; i += 1)
                {
                    auto fd = (s32)i + 1;
                    switch (options.policies[i])
                    {
                        case ExecuteStandardStreamPolicy::inherit:
                            {
                            } break;
                        case ExecuteStandardStreamPolicy::pipe:
                            {
                                close(pipes[i][0]);
                                dup2(pipes[i][1], fd);
                                close(pipes[i][1]);
                            } break;
                        case ExecuteStandardStreamPolicy::ignore:
                            {
                                dup2(null_file_descriptor, fd);
                                close(null_file_descriptor);
                            } break;
                    }
                }

                auto result = execve(arguments[0], arguments.pointer, environment.pointer);

                if (result != -1)
                {
                    unreachable();
                }

                trap();
            } break;
        default:
            {
                for (u64 i = 0; i < standard_stream_count; i += 1)
                {
                    if (options.policies[i] == ExecuteStandardStreamPolicy::pipe)
                    {
                        close(pipes[i][1]);
                    }
                }

                if (options.policies[0] == ExecuteStandardStreamPolicy::pipe || options.policies[1] == ExecuteStandardStreamPolicy::pipe)
                {
                    trap();
                }

                for (u64 i = 0; i < standard_stream_count; i += 1)
                {
                    if (options.policies[i] == ExecuteStandardStreamPolicy::pipe)
                    {
                        trap();
                    }
                }

                int status = 0;
                auto waitpid_result = waitpid(pid, &status, 0);
                
                if (waitpid_result == pid)
                {
                    trap();
                }
                else if (waitpid_result == -1)
                {
                    trap();
                }
                else
                {
                    trap();
                }
            } break;

    }

    return execution;
}
